;; Generated package description from company.el  -*- no-byte-compile: t -*-
(define-package "company" "0.10.2.0.20240911.20707" "Modular text completion framework" '((emacs "25.1")) :commit "58542350fb6a949f5ed91616a715e79ee42bd8a8" :maintainer '("Dmitry Gutov" . "dmitry@gutov.dev") :keywords '("abbrev" "convenience" "matching") :url "http://company-mode.github.io/")
